"""Center generation from data for radial basis functions."""

import abc
import logging
from typing import Any, Callable, Dict, Tuple, Union

import numpy as np
import sklearn.base
import sklearn.cluster
import sklearn.mixture
from scipy import stats

log = logging.getLogger(__name__)


class Centers(sklearn.base.BaseEstimator, metaclass=abc.ABCMeta):
    """Base class for all center generation estimators.

    All attributes with a trailing underscore must be set in the subclass'
    :func:`fit`.

    Attributes
    ----------
    centers_ : np.ndarray
        Centers, shape (n_centers, n_features).
    n_centers_ : int
        Number of centers generated.
    n_features_in_ : int
        Number of features input.
    """

    @abc.abstractmethod
    def fit(self, X: np.ndarray, y: np.ndarray = None) -> 'Centers':
        """Generate centers from data.

        Parameters
        ----------
        X : np.ndarray
            Data matrix.
        y : np.ndarray
            Ignored.

        Returns
        -------
        Centers
            Instance of itself.

        Raises
        ------
        ValueError
            If any of the constructor parameters are incorrect.
        """
        raise NotImplementedError()


class GridCenters(Centers):
    """Centers generated on a uniform grid.

    Attributes
    ----------
    centers_ : np.ndarray
        Centers, shape (n_centers, n_features).
    n_centers_ : int
        Number of centers generated.
    n_features_in_ : int
        Number of features input.
    range_max_ : np.ndarray
        Maximum value of each feature used to generate grid.
    range_min_ : np.ndarray
        Minimum value of each feature used to generate grid.

    Examples
    --------
    Generate centers on a grid

    >>> grid = pykoop.GridCenters(n_points_per_feature=4)
    >>> grid.fit(X_msd[:, 1:])  # Remove episode feature
    GridCenters(n_points_per_feature=4)
    >>> grid.centers_
    array([...])
    """

    def __init__(
        self,
        n_points_per_feature: int = 2,
        symmetric_range: bool = False,
    ) -> None:
        """Instantiate :class:`GridCenters`.

        Parameters
        ----------
        n_points_per_feature : int
            Number of points in grid for each feature.
        symmetric_range : bool
            If true, the grid range for a given feature is forced to be
            symmetric about zero (i.e., ``[-max(abs(x)), max(abs(x))]``).
            Otherwise, the grid range is taken directly on the data
            (i.e., ``[min(x), max(x)]``). Default is false.
        """
        self.n_points_per_feature = n_points_per_feature
        self.symmetric_range = symmetric_range

    def fit(self, X: np.ndarray, y: np.ndarray = None) -> 'GridCenters':
        # noqa: D102
        X = sklearn.utils.validation.check_array(X)
        self.n_features_in_ = X.shape[1]
        # Validate parameters
        if (self.n_points_per_feature is not None
                and self.n_points_per_feature <= 0):
            raise ValueError('`n_points_per_feature` must be at least one.')
        # Calculate ranges of each feature
        self.range_min_, self.range_max_ = _feature_range(
            X, self.symmetric_range)
        # Generate linspaces for each feature
        linspaces = [
            np.linspace(self.range_min_[i], self.range_max_[i],
                        self.n_points_per_feature)
            for i in range(self.n_features_in_)
        ]
        # Generate centers
        self.centers_ = np.array(np.meshgrid(*linspaces)).reshape(
            self.n_features_in_, -1).T
        self.n_centers_ = self.centers_.shape[0]
        return self


class UniformRandomCenters(Centers):
    """Centers sampled from a uniform distribution.

    Inspired by center generation approach used in [DTK20]_.

    Attributes
    ----------
    centers_ : np.ndarray
        Centers, shape (n_centers, n_features).
    n_centers_ : int
        Number of centers generated.
    n_features_in_ : int
        Number of features input.
    range_max_ : np.ndarray
        Maximum value of each feature used to generate grid.
    range_min_ : np.ndarray
        Minimum value of each feature used to generate grid.

    Examples
    --------
    Generate centers from a uniform distribution

    >>> rand = pykoop.UniformRandomCenters(n_centers=10)
    >>> rand.fit(X_msd[:, 1:])  # Remove episode feature
    UniformRandomCenters(n_centers=10)
    >>> rand.centers_
    array([...])
    """

    def __init__(
        self,
        n_centers: int = 100,
        symmetric_range: bool = False,
        random_state: Union[int, np.random.RandomState] = None,
    ) -> None:
        """Instantiate :class:`UniformRandomCenters`.

        Parameters
        ----------
        n_centers : int
            Number of centers to generate.
        symmetric_range : bool
            If true, the grid range for a given feature is forced to be
            symmetric about zero (i.e., ``[-max(abs(x)), max(abs(x))]``).
            Otherwise, the grid range is taken directly on the data
            (i.e., ``[min(x), max(x)]``). Default is false.
        random_state : Union[int, np.random.RandomState]
            Random seed.
        """
        self.n_centers = n_centers
        self.symmetric_range = symmetric_range
        self.random_state = random_state

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray = None,
    ) -> 'UniformRandomCenters':
        # noqa: D102
        X = sklearn.utils.validation.check_array(X)
        self.n_features_in_ = X.shape[1]
        # Validate parameters
        if self.n_centers <= 0:
            raise ValueError('`n_centers` must be greater than zero.')
        self.n_centers_ = self.n_centers
        # Calculate ranges of each feature
        self.range_min_, self.range_max_ = _feature_range(
            X, self.symmetric_range)
        # Generate centers in range ``[loc, loc + scale]``.
        self.centers_ = np.array([
            stats.uniform.rvs(
                loc=self.range_min_[i],
                scale=(self.range_max_[i] - self.range_min_[i]),
                size=self.n_centers_,
                random_state=self.random_state,
            ) for i in range(self.n_features_in_)
        ]).T
        return self


class GaussianRandomCenters(Centers):
    """Centers sampled from a Gaussian distribution.

    Inspired by center generation approach used in [CHH19]_.

    Attributes
    ----------
    centers_ : np.ndarray
        Centers, shape (n_centers, n_features).
    n_centers_ : int
        Number of centers generated.
    n_features_in_ : int
        Number of features input.
    mean_ : np.ndarray
        Mean feature.
    cov_ : np.ndarray
        Covariance matrix.

    Examples
    --------
    Generate centers from a Gaussian distribution

    >>> rand = pykoop.GaussianRandomCenters(n_centers=10)
    >>> rand.fit(X_msd[:, 1:])  # Remove episode feature
    GaussianRandomCenters(n_centers=10)
    >>> rand.centers_
    array([...])
    """

    def __init__(
        self,
        n_centers: int = 100,
        random_state: Union[int, np.random.RandomState] = None,
    ) -> None:
        """Instantiate :class:`GaussianRandomCenters`.

        Parameters
        ----------
        n_centers : int
            Number of centers to generate.
        random_state : Union[int, np.random.RandomState]
            Random seed.
        """
        self.n_centers = n_centers
        self.random_state = random_state

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray = None,
    ) -> 'GaussianRandomCenters':
        # noqa: D102
        X = sklearn.utils.validation.check_array(X, ensure_min_samples=2)
        self.n_features_in_ = X.shape[1]
        # Validate parameters
        if self.n_centers <= 0:
            raise ValueError('`n_centers` must be greater than zero.')
        self.n_centers_ = self.n_centers
        # Calculate mean and covariance
        self.mean_ = np.mean(X, axis=0)
        self.cov_ = np.cov(X, rowvar=False)
        # Generate centers
        self.centers_ = stats.multivariate_normal.rvs(
            mean=self.mean_,
            cov=self.cov_,
            size=self.n_centers_,
            random_state=self.random_state,
        )
        return self


class QmcCenters(Centers):
    """Centers generated with Quasi-Monte Carlo sampling.

    Attributes
    ----------
    centers_ : np.ndarray
        Centers, shape (n_centers, n_features).
    n_centers_ : int
        Number of centers generated.
    n_features_in_ : int
        Number of features input.
    range_max_ : np.ndarray
        Maximum value of each feature used to generate grid.
    range_min_ : np.ndarray
        Minimum value of each feature used to generate grid.
    qmc_ : stats.qmc.QMCEngine
        Quasi-Monte Carlo sampler instantiated from ``qmc``.

    Examples
    --------
    Generate centers using Latin hypercube sampling (default)

    >>> qmc = pykoop.QmcCenters(n_centers=10)
    >>> qmc.fit(X_msd[:, 1:])  # Remove episode feature
    QmcCenters(n_centers=10)
    >>> qmc.centers_
    array([...])

    Generate centers using a Sobol sequence

    >>> qmc = pykoop.QmcCenters(n_centers=8, qmc=scipy.stats.qmc.Sobol)
    >>> qmc.fit(X_msd[:, 1:])  # Remove episode feature
    QmcCenters(n_centers=8, qmc=<class 'scipy.stats._qmc.Sobol'>)
    >>> qmc.centers_
    array([...])
    """

    def __init__(
        self,
        n_centers: int = 100,
        symmetric_range: bool = False,
        qmc: Callable[..., stats.qmc.QMCEngine] = None,
        qmc_kw: Dict[str, Any] = None,
        random_state: Union[int, np.random.RandomState] = None,
    ) -> None:
        """Instantiate :class:`QmcCenters`.

        Parameters
        ----------
        n_centers : int
            Number of centers to generate.

        symmetric_range : bool
            If true, the grid range for a given feature is forced to be
            symmetric about zero (i.e., ``[-max(abs(x)), max(abs(x))]``).
            Otherwise, the grid range is taken directly on the data
            (i.e., ``[min(x), max(x)]``). Default is false.

        qmc : Callable[..., stats.qmc.QMCEngine]
            Quasi-Monte Carlo method from :mod:`scipy.stats.qmc` to use.
            Argument is the desired subclass of
            :class:`scipy.stats.qmc.QMCEngine` to use. Accepts the class
            itself, not an instance of the class. Possible values are

            - :class:`scipy.stats.qmc.Sobol` -- Sobol sequence,
            - :class:`scipy.stats.qmc.Halton` -- Halton sequence,
            - :class:`scipy.stats.qmc.LatinHypercube` -- Latin hypercube
              sampling (LHS),
            - :class:`scipy.stats.qmc.PoissonDisk` -- Poisson disk sampling
              (requires ``scipy`` v1.9.0.),
            - :class:`scipy.stats.qmc.MultinomialQMC` -- Multinomial
              distribution, and
            - :class:`scipy.stats.qmc.MultivariateNormalQMC` -- Multivariate
              normal distribution.

            If ``None``, defaults to Latin hypercube sampling.

        qmc_kw : Dict[str, Any]
            Additional keyword arguments passed when instantiating ``qmc``. If
            ``seed`` is specified here, it takes precedence over
            ``random_state``.

        random_state : Union[int, np.random.RandomState]
            Random seed.
        """
        self.n_centers = n_centers
        self.symmetric_range = symmetric_range
        self.qmc = qmc
        self.qmc_kw = qmc_kw
        self.random_state = random_state

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray = None,
    ) -> 'QmcCenters':
        # noqa: D102
        X = sklearn.utils.validation.check_array(X, ensure_min_samples=2)
        self.n_features_in_ = X.shape[1]
        # Validate parameters
        if self.n_centers <= 0:
            raise ValueError('`n_centers` must be greater than zero.')
        self.n_centers_ = self.n_centers
        # Calculate ranges of each feature
        self.range_min_, self.range_max_ = _feature_range(
            X, self.symmetric_range)
        # Set default QMC method
        qmc = stats.qmc.LatinHypercube if self.qmc is None else self.qmc
        # Set QMC arguments
        qmc_args = {} if self.qmc_kw is None else self.qmc_kw
        if 'seed' not in qmc_args.keys():
            qmc_args['seed'] = self.random_state
        # Instantiate QMC
        self.qmc_ = qmc(self.n_features_in_, **qmc_args)
        # Generate and scale samples
        unscaled_centers = self.qmc_.random(self.n_centers_)
        self.centers_ = stats.qmc.scale(unscaled_centers, self.range_min_,
                                        self.range_max_)
        return self


class ClusterCenters(Centers):
    """Centers generated from a clustering algorithm.

    Also supports taking centers from the means of a Gaussian mixture model.

    Inspired by center generation approach used in [DTK20]_.

    Attributes
    ----------
    centers_ : np.ndarray
        Centers, shape (n_centers, n_features).
    n_centers_ : int
        Number of centers generated.
    n_features_in_ : int
        Number of features input.
    estimator_ : sklearn.base.BaseEstimator
        Fit clustering estimator or Gaussian mixture model.

    Examples
    --------
    Generate centers using K-means clustering

    >>> kmeans = pykoop.ClusterCenters(sklearn.cluster.KMeans(n_clusters=3))
    >>> kmeans.fit(X_msd[:, 1:])  # Remove episode feature
    ClusterCenters(estimator=KMeans(n_clusters=3))
    >>> kmeans.centers_
    array([...])
    """

    def __init__(self, estimator: sklearn.base.BaseEstimator = None) -> None:
        """Instantiate :class:`ClusterCenters`.

        Parameters
        ----------
        estimator : sklearn.base.BaseEstimator
            Clustering estimator or Gaussian mixture model. Must provide
            ``cluster_centers_`` or ``means_`` once fit. Possible algorithms
            include

            - :class:`sklearn.cluster.KMeans`,
            - :class:`sklearn.cluster.AffinityPropagation`,
            - :class:`sklearn.cluster.MeanShift`,
            - :class:`sklearn.cluster.BisectingKMeans`,
            - :class:`sklearn.cluster.MiniBatchKMeans`,
            - :class:`sklearn.mixture.GaussianMixture`, or
            - :class:`sklearn.mixture.BayesianGaussianMixture`.

            The number of centers generated is controlled by the chosen
            estimator. If a random seed is desired, it must be set in the
            chosen estimator. Defaults to :class:`sklearn.cluster.KMeans`.
        """
        self.estimator = estimator

    def fit(self, X: np.ndarray, y: np.ndarray = None) -> 'ClusterCenters':
        # noqa: D102
        X = sklearn.utils.validation.check_array(X)
        self.n_features_in_ = X.shape[1]
        # Clone and fit estimator
        self.estimator_ = (sklearn.base.clone(self.estimator) if self.estimator
                           is not None else sklearn.cluster.KMeans())
        self.estimator_.fit(X)
        # Set centers
        if hasattr(self.estimator_, 'cluster_centers_'):
            self.centers_ = self.estimator_.cluster_centers_
        elif hasattr(self.estimator_, 'means_'):
            self.centers_ = self.estimator_.means_
        else:
            raise ValueError('`estimator` must provide either '
                             '`cluster_centers_` or `means_` after fit.')
        self.n_centers_ = self.centers_.shape[0]
        return self


class GaussianMixtureRandomCenters(Centers):
    """Centers generated from sampling a Gaussian mixture model.

    Attributes
    ----------
    centers_ : np.ndarray
        Centers, shape (n_centers, n_features).
    n_centers_ : int
        Number of centers generated.
    n_features_in_ : int
        Number of features input.
    estimator_ : sklearn.base.BaseEstimator
        Fit Gaussian mixture model.

    Examples
    --------
    Generate centers by sampling a Gaussian mixture model

    >>> gmm = pykoop.GaussianMixtureRandomCenters(n_centers=100,
    ...     estimator=sklearn.mixture.GaussianMixture(n_components=3))
    >>> gmm.fit(X_msd[:, 1:])  # Remove episode feature
    GaussianMixtureRandomCenters(estimator=GaussianMixture(n_components=3))
    >>> gmm.centers_
    array([...])
    """

    def __init__(
        self,
        n_centers: int = 100,
        estimator: sklearn.base.BaseEstimator = None,
    ) -> None:
        """Instantiate :class:`GaussianMixtureRandomCenters`.

        Parameters
        ----------
        n_centers : int
            Number of centers to generate.

        estimator : sklearn.base.BaseEstimator
            Gaussian mixture model. Possible algorithms include

            - :class:`sklearn.mixture.GaussianMixture`, or
            - :class:`sklearn.mixture.BayesianGaussianMixture`.

            If a random seed is desired, it must be set in the
            chosen estimator. Defaults to
            :class:`sklearn.mixture.GaussianMixture(n_components=2)`.
        """
        self.n_centers = n_centers
        self.estimator = estimator

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray = None,
    ) -> 'GaussianMixtureRandomCenters':
        # noqa: D102
        X = sklearn.utils.validation.check_array(X)
        self.n_features_in_ = X.shape[1]
        # Clone and fit estimator
        self.estimator_ = (sklearn.base.clone(self.estimator) if self.estimator
                           is not None else sklearn.mixture.GaussianMixture(
                               n_components=2))
        self.estimator_.fit(X)
        # Sample fit distribution
        self.centers_ = self.estimator_.sample(self.n_centers)[0]
        self.n_centers_ = self.centers_.shape[0]
        return self


class DataCenters(Centers):
    """Centers taken from raw data upon instantiation or fit.

    Attributes
    ----------
    centers_ : np.ndarray
        Centers, shape (n_centers, n_features).
    n_centers_ : int
        Number of centers generated.
    n_features_in_ : int
        Number of features input.
    """

    def __init__(self, centers: np.ndarray = None):
        """Instantiate :class:`DataCenters`.

        Parameters
        ----------
        centers : np.ndarray
            Array of centers, shape (n_centers, n_features). If ``None``, then
            centers are taken from fit.
        """
        self.centers = centers

    def fit(self, X: np.ndarray, y: np.ndarray = None) -> 'DataCenters':
        # noqa: D102
        X = sklearn.utils.validation.check_array(X)
        self.n_features_in_ = X.shape[1]
        # Set centers
        if self.centers is None:
            self.centers_ = X
        else:
            if self.centers.shape[1] != self.n_features_in_:
                raise ValueError('`centers` must have shape '
                                 '(n_centers, n_features).')
            self.centers_ = self.centers
        # Set center shape
        self.n_centers_ = self.centers_.shape[0]
        return self


def _feature_range(
    X: np.ndarray,
    symmetric_range: bool,
) -> Tuple[np.ndarray, np.ndarray]:
    """Get range of features from data matrix.

    Parameters
    ----------
    X : np.ndarray
        Data matrix.
    symmetric_range : bool
        If true, the grid range for a given feature is forced to be symmetric
        about zero (i.e., ``[-max(abs(x)), max(abs(x))]``). Otherwise, the grid
        range is taken directly on the data (i.e., ``[min(x), max(x)]``).

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Minumum and maximum values of each feature.
    """
    if symmetric_range:
        range_max = np.max(np.abs(X), axis=0)
        range_min = -range_max
    else:
        range_max = np.max(X, axis=0)
        range_min = np.min(X, axis=0)
    return (range_min, range_max)



"""Kernel approximations for corresponding lifting functions."""

import abc
import logging
from typing import Dict, List, Tuple, Union, Any

import numpy as np
import scipy.stats
import sklearn.base
import sklearn.preprocessing
import sklearn.utils

# Create logger
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class KernelApproximation(
        sklearn.base.BaseEstimator,
        sklearn.base.TransformerMixin,
        metaclass=abc.ABCMeta,
):
    """Base class for all kernel approximations.

    All attributes with a trailing underscore must be set in the subclass'
    :func:`fit`.

    Attributes
    ----------
    n_features_in_ : int
        Number of features input.
    n_features_out_ : int
        Number of features output. This attribute is not available in
        estimators from :mod:`sklearn.kernel_approximation`.
    """

    @abc.abstractmethod
    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray = None,
    ) -> 'KernelApproximation':
        """Fit kernel approximation.

        Parameters
        ----------
        X : np.ndarray
            Data matrix.
        y : np.ndarray
            Ignored.

        Returns
        -------
        KernelApproximation
            Instance of itself.

        Raises
        ------
        ValueError
            If any of the constructor parameters are incorrect.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def transform(self, X: np.ndarray) -> np.ndarray:
        """Transform data.

        Parameters
        ----------
        X : np.ndarray
            Data matrix.

        Returns
        -------
        np.ndarray
            Transformed data matrix.
        """
        raise NotImplementedError()


class RandomFourierKernelApprox(KernelApproximation):
    """Kernel approximation with random Fourier features.

    Similar to :class:`sklearn.kernel_approximation.RBFSampler`, but supports
    kernels other than just the Gaussian.

    For details, see [RR07]_.

    Attributes
    ----------
    n_features_in_ : int
        Number of features input.
    n_features_out_ : int
        Number of features output. This attribute is not available in
        estimators from :mod:`sklearn.kernel_approximation`.
    ift_ : scipy.stats.rv_continuous
        Probability distribution corresponding to inverse Fourier transform of
        chosen kernel.
    random_weights_ : np.ndarray, shape (n_features, n_components)
        Random weights to inner-product with features.
    random_offsets_ : np.ndarray, shape (n_features, )
        Random offsets if ``method`` is ``'weight_offset'``.

    Examples
    --------
    Generate random Fourier features from a Gaussian kernel

    >>> ka = pykoop.RandomFourierKernelApprox(
    ...     kernel_or_ift='gaussian',
    ...     n_components=10,
    ...     shape=1,
    ...     random_state=1234,
    ... )
    >>> ka.fit(X_msd[:, 1:])  # Remove episode feature
    RandomFourierKernelApprox(n_components=10, random_state=1234)
    >>> ka.transform(X_msd[:, 1:])
    array([...])
    """

    _ift_lookup = {
        'gaussian': scipy.stats.norm,
        'laplacian': scipy.stats.cauchy,
        'cauchy': scipy.stats.laplace,
    }
    """Lookup table for inverse Fourier transform of kernel.

    Laplacian and Cauchy being swapped is not a typo. They are Fourier
    transforms of each other.
    """

    def __init__(
        self,
        kernel_or_ift: Union[str, scipy.stats.rv_continuous] = 'gaussian',
        n_components: int = 100,
        shape: float = 1,
        method: str = 'weight_offset',
        random_state: Union[int, np.random.RandomState] = None,
    ) -> None:
        """Instantiate :class:`RandomFourierKernelApprox`.

        Parameters
        ----------
        kernel_or_ift : Union[str, scipy.stats.rv_continuous]
            Kernel to approximate. Possible options are

                - ``'gaussian'`` -- Gaussian kernel, with inverse Fourier
                  transform :class:`scipy.stats.norm` (default),
                - ``'laplacian'`` -- Laplacian kernel, with inverse Fourier
                  transform :class:`scipy.stats.cauchy`, or
                - ``'cauchy'`` -- Cauchy kernel, with inverse Fourier transform
                  :class:`scipy.stats.laplace`.

            Alternatively, a positive, shift-invariant kernel can be implicitly
            specified by providing its inverse Fourier transform as a
            univariate probability distribution subclassing
            :class:`scipy.stats.rv_continuous`.

        n_components : int
            Number of random samples used to generate features. If
            ``method='weight_offset'``, this corresponds directly to the number
            of features generated. If ``method='weight_only'``, then
            ``2 * n_components`` features are generated.

        shape : float
            Shape parameter. Must be greater than zero. Larger numbers
            correspond to "sharper" kernels. Scaled to be consistent with
            ``gamma`` from :class:`sklearn.kernel_approximation.RBFSampler`.
            This can lead to a mysterious factor of ``sqrt(2)`` in other
            kernels. Default is ``1``.

        method : str
            Feature generation method to use. If ``'weight_offset'`` (default),
            each weight corresponds to a feature
            ``cos(weight.T @ x + offset)``. If ``'weight_only'``, then each
            weight corresponds to two features, ``cos(weight.T @ x)`` and
            ``sin(weight.T @ x)``, meaning the number of features generated is
            ``2 * n_components``.

        random_state : Union[int, np.random.RandomState]
            Random seed.
        """
        self.kernel_or_ift = kernel_or_ift
        self.n_components = n_components
        self.shape = shape
        self.method = method
        self.random_state = random_state

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray = None,
    ) -> 'RandomFourierKernelApprox':
        """Fit kernel approximation.

        Parameters
        ----------
        X : np.ndarray
            Data matrix.
        y : np.ndarray
            Ignored.

        Returns
        -------
        RandomFourierKernelApprox
            Instance of itself.

        Raises
        ------
        ValueError
            If any of the constructor parameters are incorrect.
        """
        X = sklearn.utils.validation.check_array(X)
        # Set inverse Fourier transform
        if isinstance(self.kernel_or_ift, str):
            self.ift_ = self._ift_lookup[self.kernel_or_ift]
        else:
            self.ift_ = self.kernel_or_ift
        # Validate input
        if self.n_components <= 0:
            raise ValueError('`n_components` must be positive.')
        if self.shape <= 0:
            raise ValueError('`shape` must be positive.')
        valid_methods = ['weight_offset', 'weight_only']
        if self.method not in valid_methods:
            raise ValueError(f'`method` must be one of {valid_methods}.')
        # Set number of input and output features
        self.n_features_in_ = X.shape[1]
        if self.method == 'weight_only':
            self.n_features_out_ = 2 * self.n_components
        else:
            self.n_features_out_ = self.n_components
        # Generate random weights
        self.random_weights_ = self.ift_.rvs(
            scale=1,
            size=(self.n_features_in_, self.n_components),
            random_state=self.random_state,
        )
        # Generate random offsets if needed
        if self.method == 'weight_only':
            self.random_offsets_ = None
        else:
            # Range is [loc, loc + scale]
            self.random_offsets_ = scipy.stats.uniform.rvs(
                loc=0,
                scale=(2 * np.pi),
                size=self.n_components,
                random_state=self.random_state,
            )
        # Easiest way to make sure distribution is univariate is to check the
        # dimension of the output.
        if self.random_weights_.ndim != 2:
            raise ValueError('`kernel_or_ift` must specify a univariate '
                             'probability distribution.')
        return self

    def transform(self, X: np.ndarray) -> np.ndarray:
        """Transform data.

        Parameters
        ----------
        X : np.ndarray
            Data matrix.

        Returns
        -------
        np.ndarray
            Transformed data matrix.
        """
        sklearn.utils.validation.check_is_fitted(self)
        X = sklearn.utils.validation.check_array(X)
        X_scaled = np.sqrt(2 * self.shape) * X
        products = X_scaled @ self.random_weights_  # (n_samples, n_components)
        if self.method == 'weight_only':
            Xt_unscaled = np.hstack((
                np.cos(products),
                np.sin(products),
            ))
        else:
            Xt_unscaled = np.sqrt(2) * np.cos(products + self.random_offsets_)
        Xt = np.sqrt(1 / self.n_components) * Xt_unscaled
        return Xt


class RandomBinningKernelApprox(KernelApproximation):
    r"""Kernel approximation with random binning.

    Highly experimental! For more details, see [RR07]_.

    Attributes
    ----------
    n_features_in_ : int
        Number of features input.
    n_features_out_ : int
        Number of features output. This attribute is not available in
        estimators from :mod:`sklearn.kernel_approximation`.
    ddot_ : scipy.stats.rv_continuous
        Probability distribution corresponding to ``\delta \ddot{k}(\delta)``.
    pitches_ : np.ndarray, shape (n_features, n_components)
        Grid pitches for each component.
    shifts_ : np.ndarray, shape (n_features, n_components)
        Grid shifts for each component.
    encoder_ : sklearn.preprocessing.OneHotEncoder
        One-hot encoder used for hashing sample coordinates for each component.

    Examples
    --------
    Generate randomly binned features from a Laplacian kernel

    >>> ka = pykoop.RandomBinningKernelApprox(
    ...     kernel_or_ddot='laplacian',
    ...     n_components=10,
    ...     shape=1,
    ...     random_state=1234,
    ... )
    >>> ka.fit(X_msd[:, 1:])  # Remove episode feature
    RandomBinningKernelApprox(n_components=10, random_state=1234)
    >>> ka.transform(X_msd[:, 1:])
    array([...])
    """

    _ddot_lookup = {
        'laplacian': scipy.stats.gamma(a=2),
    }
    r"""Lookup table for ``\delta \ddot{k}(\delta)``."""

    def __init__(
        self,
        kernel_or_ddot: Union[str, scipy.stats.rv_continuous] = 'laplacian',
        n_components: int = 100,
        shape: float = 1,
        encoder_kwargs: Dict[str, Any] = None,
        random_state: Union[int, np.random.RandomState] = None,
    ) -> None:
        r"""Instantiate :class:`RandomBinningKernelApprox`.

        Parameters
        ----------
        kernel_or_ddot : Union[str, scipy.stats.rv_continuous]
            Kernel to approximate. Possible options are

                - ``'laplacian'`` -- Laplacian kernel, with
                  ``\delta \ddot{k}(\delta)`` being :class:`scipy.stats.gamma`
                  with shape parameter ``a=2`` (default).

            Alternatively, a separable, positive, shift-invariant kernel can be
            implicitly specified by providing ``\delta \ddot{k}(\delta)`` as a
            univariate probability distribution subclassing
            :class:`scipy.stats.rv_continuous`.

        n_components : int
            Number of random samples used to generate features. The higher the
            number of components, the higher the number of features. Since
            unoccupied bins are eliminated, it's impossible to know the exact
            number of features before fitting.

        shape : float
            Shape parameter. Must be greater than zero. Larger numbers
            correspond to "sharper" kernels. Scaled to be consistent with
            ``gamma`` from :class:`sklearn.kernel_approximation.RBFSampler`.
            This can lead to a mysterious factor of ``sqrt(2)`` in other
            kernels. Default is ``1``.

        encoder_kwargs : Dict[str, Any]
            Extra keyword arguments for internal
            :class:`sklearn.preprocessing.OneHotEncoder`. For experimental use
            only. The wrong arguments can break everything. Overrides defaults.

        random_state : Union[int, np.random.RandomState]
            Random seed.
        """
        self.kernel_or_ddot = kernel_or_ddot
        self.n_components = n_components
        self.shape = shape
        self.encoder_kwargs = encoder_kwargs
        self.random_state = random_state

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray = None,
    ) -> 'RandomBinningKernelApprox':
        """Fit kernel approximation.

        Parameters
        ----------
        X : np.ndarray
            Data matrix.
        y : np.ndarray
            Ignored.

        Returns
        -------
        RandomBinningKernelApprox
            Instance of itself.

        Raises
        ------
        ValueError
            If any of the constructor parameters are incorrect.
        """
        X = sklearn.utils.validation.check_array(X)
        X_scaled = np.sqrt(2 * self.shape) * X
        # Set ``ddot_``
        if isinstance(self.kernel_or_ddot, str):
            self.ddot_ = self._ddot_lookup[self.kernel_or_ddot]
        else:
            self.ddot_ = self.kernel_or_ddot
        # Validate input
        if self.n_components <= 0:
            raise ValueError('`n_components` must be positive.')
        if self.shape <= 0:
            raise ValueError('`shape` must be positive.')
        # Set number of input and output features
        self.n_features_in_ = X_scaled.shape[1]
        # Generate grids
        self.pitches_ = self.ddot_.rvs(
            size=(self.n_features_in_, self.n_components),
            random_state=self.random_state,
        )
        self.shifts_ = scipy.stats.uniform.rvs(
            loc=0,
            scale=self.pitches_,
            random_state=self.random_state,
        )
        # Hash samples and fit one-hot encoder
        X_hashed = self._hash_samples(X_scaled)
        encoder_args = {
            'categories': 'auto',
            'sparse': False,
            'handle_unknown': 'ignore',
        }
        if self.encoder_kwargs is not None:
            encoder_args.update(self.encoder_kwargs)
        self.encoder_ = sklearn.preprocessing.OneHotEncoder(**encoder_args)
        self.encoder_.fit(X_hashed)
        # Get number of output features from the encoder
        self.n_features_out_ = np.concatenate(self.encoder_.categories_).size
        return self

    def transform(self, X: np.ndarray) -> np.ndarray:
        """Transform data.

        Parameters
        ----------
        X : np.ndarray
            Data matrix.

        Returns
        -------
        np.ndarray
            Transformed data matrix.
        """
        sklearn.utils.validation.check_is_fitted(self)
        X = sklearn.utils.validation.check_array(X)
        X_scaled = np.sqrt(2 * self.shape) * X
        X_hashed = self._hash_samples(X_scaled)
        Xt = self.encoder_.transform(X_hashed) / np.sqrt(self.n_components)
        return Xt

    def _hash_samples(self, X: np.ndarray) -> np.ndarray:
        """Calculate bin coordinates of each sample and hash them.

        Parameters
        ----------
        X : np.ndarray, shape (n_samples, n_features)
            Data matrix.

        Returns
        -------
        np.ndarray, shape (n_samples, n_components)
            Hashed bin coordinates for each component.
        """
        # Calculate bin coordinates of each sample
        # Shape is (n_samples, n_features, n_components)
        coords = np.floor((X[:, :, np.newaxis] - self.shifts_) / self.pitches_)
        # Convert each coordinate to tuple (so it's immutable), then compute
        # its hash.
        coords_hashed = np.array(
            [[
                hash(tuple(coords[sample, :, component].astype(int)))
                for component in range(coords.shape[2])
            ] for sample in range(coords.shape[0])],
            dtype=int,
        )
        return coords_hashed
